/*
 * cbd.c
 * kirk johnson
 * january 1993
 *
 * define PATCH_BROKEN_DATA to enable hand-hacked patch to fix what
 * appear to be some broken elements of the database
 *
 * RCS $Id: cbd.c,v 1.4 1994/05/13 20:32:55 tuna Exp $
 *
 * Copyright 1993 by Kirk Lauritz Johnson (see the included
 * file "kljcpyrt.h" for complete copyright information)
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cbd.h"
#include "kljcpyrt.h"

#define SHORTFLAG (0x4000)
#define CBD_MAGIC (0x20770002)

#ifdef sun
#define SWAP_BYTES
#endif

#define PATCH_BROKEN_DATA

#ifdef PATCH_BROKEN_DATA

typedef struct
{
  int     idx;			/* seg idx          */
  int     length;		/* seg length       */
  CBD_POS head[3];		/* first 3 seg data */
  int     from;			/* from idx         */
  int     to;			/* to idx           */
  CBD_POS patch;		/* patch value      */
} DataPatch;

#define NUM_DATA_PATCHES (3)

DataPatch patch[NUM_DATA_PATCHES] =
{
{
  8378, 1361,
  643321, -61209, 643308, -61201, 643295, -61188,
  615, 1360,
  14720, 0,
},
{
  8464, 178,
  647707, -61197, 647696, -61181, 647688, -61164,
  161, 177,
  14720, 0,
},
{
  2399, 540,
  643649, 254876, 643735, 254879, 643806, 254892,
  393, 539,
  14720, 0,
},
};

#endif /* PATCH_BROKEN_DATA */

typedef struct
{
  int magic;			/* magic number                 */
  int dictaddr;			/* offset of segment dictionary */
  int segcount;			/* number of segments           */
  int segsize;			/* size of segment dictionary   */
  int segmax;			/* max # strokes in segment (?) */
  int fill[(40/sizeof(int))-5]; /* junk                         */
} CBDhead;

typedef struct
{
  int   segid;                  /* segment id           */
  int   maxlat, minlat;		/* max, min lat         */
  int   maxlong, minlong;	/* max, min long        */
  int   absaddr;                /* abs addr of seg data */
  short nbytes;                 /* number of data bytes */
  short rank;                   /* segment rank         */
} SegDict;

typedef struct
{
  int   orgx;                   /* starting x */
  int   orgy;                   /* starting y */
  int   id;                     /* segment id */
  short nstrokes;               /* # strokes  */
  short dummy;                  /* junk       */
} Segment;


short read16(s)
     FILE *s;
{
  short rslt;

  if (fread(&rslt, sizeof(short), 1, s) != 1)
    cbd_fatal("read16", "unable to read");

  return rslt;
}


#ifdef SWAP_BYTES

static short swap16(x)
     short x;
{
  short rslt;

  rslt   = x & 0xff;
  rslt <<= 8;
  x    >>= 8;
  rslt  |= x & 0xff;

  return rslt;
}

static int swap32(x)
     int x;
{
  int rslt;

  rslt   = x & 0xff;
  rslt <<= 8;
  x    >>= 8;
  rslt  |= x & 0xff;
  rslt <<= 8;
  x    >>= 8;
  rslt  |= x & 0xff;
  rslt <<= 8;
  x    >>= 8;
  rslt  |= x & 0xff;

  return rslt;
}

#else

#define swap16(x) (x)
#define swap32(x) (x)

#endif


CBD_FILE *cbd_open_file(fname)
     char *fname;
{
  int       i;
  FILE     *s;
  CBDhead   header;
  SegDict   segdict;
  Segment   seg;
  CBD_SEG  *dict;
  CBD_FILE *rslt;

  /* open input file */
  s = fopen(fname, "r");
  if (s == NULL)
    cbd_fatal("cbd_open_file", "unable to open input file");

  /* read in the file header */
  if (fread(&header, sizeof(CBDhead), 1, s) != 1)
    cbd_fatal("cbd_open_file", "unable to read header");

  /* fix byte ordering in header */
  header.magic    = swap32(header.magic);
  header.dictaddr = swap32(header.dictaddr);
  header.segcount = swap32(header.segcount);
  header.segsize  = swap32(header.segsize);
  header.segmax   = swap32(header.segmax);

  /* check the magic number */
  if (header.magic != CBD_MAGIC)
    cbd_fatal("cbd_open_file", "bad magic number");

  /* allocate space for dict */
  dict = (CBD_SEG *) malloc(sizeof(CBD_SEG) * header.segcount);
  if (dict == NULL)
    cbd_fatal("cbd_open_file", "unable to malloc");

  /* locate segment dictionary */
  if (fseek(s, header.dictaddr, 0) == -1)
    cbd_fatal("cbd_open_file", "unable to seek dictionary");

  /* scan thru segment dictionary filling in dict */
  for (i=0; i<header.segcount; i++)
  {
    if (fread(&segdict, sizeof(SegDict), 1, s) != 1)
      cbd_fatal("cbd_open_file", "unable to read segdict");

    dict[i].addr = swap32(segdict.absaddr);
    dict[i].rank = swap16(segdict.rank);
  }

  /* scan thru segments getting size of each */
  for (i=0; i<header.segcount; i++)
  {
    /* seek start of segment */
    if (fseek(s, dict[i].addr, 0) == -1)
      cbd_fatal("cbd_open_file", "unable to seek seg");
    
    /* read seg header */
    if (fread(&seg, sizeof(Segment), 1, s) != 1)
      cbd_fatal("cbd_open_file", "unable to read segment");

    dict[i].size = swap16(seg.nstrokes) + 1;
  }

  /* allocate result struct */
  rslt = (CBD_FILE *) malloc(sizeof(CBD_FILE));
  if (rslt == NULL)
    cbd_fatal("cbd_open_file", "unable to malloc");

  rslt->s     = s;
  rslt->nsegs = header.segcount;
  rslt->dict  = dict;

  return rslt;
}


void cbd_close_file(cbd)
     CBD_FILE *cbd;
{
  free(cbd->dict);
  fclose(cbd->s);
  free(cbd);
}


CBD_POS *cbd_extract_segment(cbd, idx, buf)
     CBD_FILE *cbd;
     int       idx;
     CBD_POS  *buf;
{
  int      i, j;
  CBD_SEG *dict;
  int      size;
  Segment  seg;
  int      next;
  int      lon, lat;
  int      dx, dy;

  dict = cbd->dict;
  size = dict[idx].size;

  if (buf == NULL)
  {
    buf = (CBD_POS *) malloc(sizeof(CBD_POS) * size);
    if (buf == NULL)
      cbd_fatal("cbd_extract_segment", "unable to malloc");
  }

  /* seek start of segment */
  if (fseek(cbd->s, dict[idx].addr, 0) == -1)
    cbd_fatal("cbd_extract_segment", "unable to seek seg");

  /* read seg header */
  if (fread(&seg, sizeof(Segment), 1, cbd->s) != 1)
    cbd_fatal("cbd_extract_segment", "unable to read segment");

  lon        = swap32(seg.orgx);
  lat        = swap32(seg.orgy);
  buf[0].lon = lon;
  buf[0].lat = lat;

  for (i=1; i<size; i++)
  {
    next = swap16(read16(cbd->s));
    if (next & SHORTFLAG)
    {
      /* flag bit on: unpack a 16-bit field into dx and dy */
      if (next > 0)
	next &= ~SHORTFLAG;
      dy = next & 0xFF;
      if (dy & 0x80)
	dy |= ~0xFF;         /* extend sign */
      dx = next >> 8;
      if (dx & 0x80)
	dx |= ~0xBF;         /* extend sign */
      }
    else
    {
      /* flag bit off: take dx and dy from 32-bit fields */
      if (next < 0)
	next |= SHORTFLAG;
      dx = (next << 16) | swap16(read16(cbd->s));
      next = swap16(read16(cbd->s));
      if (next < 0)
	next |= SHORTFLAG;
      dy = (next << 16) | swap16(read16(cbd->s));
    }

    lon       += dx;
    lat       += dy;
    buf[i].lon = lon;
    buf[i].lat = lat;
  }

#ifdef PATCH_BROKEN_DATA
  for (i=0; i<NUM_DATA_PATCHES; i++)
    if ((patch[i].idx == idx) &&
	(patch[i].length == size) &&
	(patch[i].head[0].lon == buf[0].lon) &&
	(patch[i].head[0].lat == buf[0].lat) &&
	(patch[i].head[1].lon == buf[1].lon) &&
	(patch[i].head[1].lat == buf[1].lat) &&
	(patch[i].head[2].lon == buf[2].lon) &&
	(patch[i].head[2].lat == buf[2].lat))
    {
      for (j=patch[i].from; j<=patch[i].to; j++)
      {
	buf[j].lon += patch[i].patch.lon;
	buf[j].lat += patch[i].patch.lat;
      }
    }
#endif /* PATCH_BROKEN_DATA */

  return buf;
}


void cbd_pos_from_lon_lat(lon, lat, pos)
     int     lon;
     int     lat;
     double *pos;
{
  double  a, b, c;

  a = lon * ((M_PI/180)/3600);
  b = lat * ((M_PI/180)/3600);
  c = cos(b);

  pos[0] = cos(a) * c;
  pos[1] = sin(a) * c;
  pos[2] = sin(b);
}


void cbd_lon_lat_from_pos(pos, lon, lat)
     double *pos;
     int    *lon;
     int    *lat;
{
  double a, b, c;

  b = asin(pos[2]);
  c = cos(b);
  a = (c == 0) ? 0 : atan2(pos[1], pos[0]);

  *lon = rint(a * (3600/(M_PI/180)));
  *lat = rint(b * (3600/(M_PI/180)));
}


void cbd_fatal(proc, msg)
     char *proc;
     char *msg;
{
  fflush(stdout);
  fprintf(stderr, "\n%s(): %s\n", proc, msg);
  exit(1);
}
