/*
 * cbdinfo.c
 * kirk johnson
 * february 1993
 *
 * collect and print information about segments
 * in a cbd file, broken down by rank
 *
 * RCS $Id: cbdinfo.c,v 1.2 1994/05/13 20:31:09 tuna Exp $
 *
 * Copyright 1993 by Kirk Lauritz Johnson (see the included
 * file "kljcpyrt.h" for complete copyright information)
 */

#include <stdio.h>
#include <math.h>
#include "cbd.h"
#include "kljcpyrt.h"

#define MAX_RANK (20)

unsigned long   seg_count[MAX_RANK+1];
unsigned long   stroke_count[MAX_RANK+1];
double          seg_length[MAX_RANK+1];


main(argc, argv)
     int   argc;
     char *argv[];
{
  int i;

  for (i=1; i<argc; i++)
    scan_file(argv[i]);

  return 0;
}


scan_file(fname)
     char *fname;
{
  int       i, j;
  int       size, rank;
  CBD_FILE *cbd;
  CBD_POS  *seg;
  double    prev[3];
  double    curr[3];
  double    delta, dist;
  double    length;

  /* open cbd file */
  cbd = cbd_open_file(fname);

  /* clear stats */
  for (i=0; i<=MAX_RANK; i++)
  {
    seg_count[i]    = 0;
    stroke_count[i] = 0;
    seg_length[i]   = 0;
  }

  /* examine each segment */
  for (i=0; i<cbd->nsegs; i++)
  {
    size = cbd->dict[i].size;
    rank = cbd->dict[i].rank;
    seg  = cbd_extract_segment(cbd, i, NULL);

    length = 0;
    cbd_pos_from_lon_lat(seg[0].lon, seg[0].lat, prev);
    for (j=1; j<size; j++)
    {
      cbd_pos_from_lon_lat(seg[j].lon, seg[j].lat, curr);

      delta   = curr[0] - prev[0];
      dist    = delta * delta;
      delta   = curr[1] - prev[1];
      dist   += delta * delta;
      delta   = curr[2] - prev[2];
      dist   += delta * delta;
      length += sqrt(dist);

      prev[0] = curr[0];
      prev[1] = curr[1];
      prev[2] = curr[2];
    }

    seg_count[rank]    += 1;
    stroke_count[rank] += size - 1;
    seg_length[rank]   += 25000 * (length / (2*M_PI));

    free(seg);
  }

  cbd_close_file(cbd);

  printf("%s:\n", fname);
  printf("  rank  # segs  strokes/seg  miles/stroke\n");
  printf("  ----  ------  -----------  ------------\n");
  for (i=0; i<=MAX_RANK; i++)
    if (seg_count[i] > 0)
      printf("  %4d  %6lu  %11.1lf  %12.8lf\n",
	     i, seg_count[i],
	     (double) stroke_count[i] / seg_count[i],
	     seg_length[i] / stroke_count[i]);
  printf("\n");
}
